"""
Fake TCP Server to process HTTP transport requests.

Instead of listening for incoming connections, it connects to Exasol and uses proxy magic.
This approach allows to bypass various connectivity problems (e.g. firewall).
"""

import os
import socket
import socketserver
import struct
import sys
import zlib


class ExaTCPServer(socketserver.TCPServer):
    timeout = 5

    def __init__(self, *args, **kwargs):
        self.proxy_host = None
        self.proxy_port = None

        self.compression = kwargs.pop('compression', False)
        self.encryption = kwargs.pop('encryption', False)

        self.total_clients = 0

        super().__init__(*args, **kwargs)

    def server_bind(self):
        self.set_sock_opts()

        """ Special Exasol packet to establish tunneling and return proxy host and port which can be used in query """
        self.socket.connect(self.server_address)
        self.socket.sendall(struct.pack("iii", 0x02212102, 1, 1))
        _, port, host = struct.unpack("ii16s", self.socket.recv(24))

        self.proxy_host = host.replace(b'\x00', b'').decode()
        self.proxy_port = port

        if self.encryption:
            context = self.generate_adhoc_ssl_context()
            self.socket = context.wrap_socket(self.socket, server_side=True, do_handshake_on_connect=False)

    def server_activate(self): pass

    def get_request(self):
        return self.socket, self.server_address

    def shutdown_request(self, request): pass

    def close_request(self, request): pass

    def set_sock_opts(self):
        # only large packets are expected for HTTP transport
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        # default keep-alive once a minute, 5 probes
        keepidle = 60
        keepintvl = 60
        keepcnt = 5

        if sys.platform.startswith("linux"):
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, keepidle)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, keepintvl)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, keepcnt)

        elif sys.platform.startswith("darwin"):
            # TCP_KEEPALIVE = 0x10
            # https://bugs.python.org/issue34932
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.socket.setsockopt(socket.IPPROTO_TCP, 0x10, keepidle)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, keepintvl)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, keepcnt)

        elif sys.platform.startswith("win"):
            self.socket.ioctl(socket.SIO_KEEPALIVE_VALS, (1, keepidle * 1000, keepintvl * 1000))

    @staticmethod
    def check_orphaned(initial_ppid):
        """
        Raise exception if current process is "orphaned" (parent process is dead)
        It is useful to stop PyEXASOL HTTP servers from being stuck in process list after parent process was killed

        Currently it works only for POSIX OS
        Please let me know if you know a good way to detect orphans on Windows
        """
        if sys.platform.startswith("win"):
            return

        current_ppid = os.getppid()

        if initial_ppid and current_ppid != initial_ppid:
            raise RuntimeError(f"Current process is orphaned, initial ppid={initial_ppid}, current ppid={current_ppid}")

    @staticmethod
    def generate_adhoc_ssl_context():
        """
        Create temporary self-signed certificate for encrypted HTTP transport
        Exasol does not check validity of certificates
        """
        from OpenSSL import crypto

        import pathlib
        import ssl
        import tempfile

        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, 2048)

        cert = crypto.X509()
        cert.set_serial_number(1)
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(60 * 60 * 24 * 365)

        cert.set_pubkey(k)
        cert.sign(k, 'sha256')

        # TemporaryDirectory is used instead of NamedTemporaryFile for compatibility with Windows
        with tempfile.TemporaryDirectory(prefix='pyexasol_ssl_') as tempdir:
            tempdir = pathlib.Path(tempdir)

            cert_file = open(tempdir / 'cert', 'wb')
            cert_file.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
            cert_file.close()

            key_file = open(tempdir / 'key', 'wb')
            key_file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k))
            key_file.close()

            context = ssl.SSLContext()
            context.verify_mode = ssl.CERT_NONE
            context.load_cert_chain(certfile=cert_file.name, keyfile=key_file.name)

            return context


class ExaHTTPRequestHandler(socketserver.StreamRequestHandler):
    """
    Oversimplified request handler to reduce "import" overhead
    """
    default_headers = b'HTTP/1.1 200 OK\r\n' \
                      b'Connection: close\r\n' \
                      b'\r\n'

    def handle(self):
        self.server.total_clients += 1

        # Extract method from the first header
        method = str(self.rfile.readline(), 'iso-8859-1').split()[0]

        # Skip all other headers
        while self.rfile.readline() != b'\r\n':
            pass

        if method == 'PUT':
            self.method_put()
        else:
            self.method_get()

    def method_put(self):
        # Compressed data loop
        if self.server.compression:
            d = zlib.decompressobj(wbits=16 + zlib.MAX_WBITS)

            while True:
                data = self.read_chunk()

                if data is None:
                    sys.stdout.buffer.write(d.flush())
                    break

                sys.stdout.buffer.write(d.decompress(data))

        # Normal data loop
        else:
            while True:
                data = self.read_chunk()

                if data is None:
                    break

                sys.stdout.buffer.write(data)

        sys.stdout.buffer.close()

        self.wfile.write(self.default_headers)

    def read_chunk(self):
        hex_length = self.rfile.readline().rstrip()

        if len(hex_length) == 0:
            chunk_len = 0
        else:
            chunk_len = int(hex_length, 16)

        if chunk_len == 0:
            return None

        data = self.rfile.read(chunk_len)

        if self.rfile.read(2) != b'\r\n':
            raise RuntimeError('Got wrong chunk delimiter in HTTP')

        return data

    def method_get(self):
        self.wfile.write(self.default_headers)

        # Compressed data loop
        if self.server.compression:
            c = zlib.compressobj(level=1, wbits=16 + zlib.MAX_WBITS)

            while True:
                data = sys.stdin.buffer.read(65536)

                if data is None or len(data) == 0:
                    self.wfile.write(c.flush(zlib.Z_FINISH))
                    break

                self.wfile.write(c.compress(data))

        # Normal data loop
        else:
            while True:
                data = sys.stdin.buffer.read(65536)

                if data is None or len(data) == 0:
                    break

                self.wfile.write(data)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(prog='python -m pyexasol_utils.http_transport'
                                     , description='TCP server for PyEXASOL HTTP transport')

    parser.add_argument('--host', help='Exasol host')
    parser.add_argument('--port', help='Exasol port', type=int)
    parser.add_argument('--mode', help='EXPORT or IMPORT')
    parser.add_argument('--ppid', help='PID of parent process', default=0, type=int)
    parser.add_argument('--compression', default=False, help='Enable compression', action='store_true')
    parser.add_argument('--encryption', default=False, help='Enable encryption', action='store_true')

    args = parser.parse_args()

    # Start TCP server
    server = ExaTCPServer((args.host, args.port)
                          , ExaHTTPRequestHandler
                          , compression=args.compression, encryption=args.encryption)

    # Send proxy string to the main process
    sys.stdout.buffer.write(f'{server.proxy_host}:{server.proxy_port}\n'.encode())
    sys.stdout.buffer.flush()

    # Handle exactly one connection
    while server.total_clients == 0:
        server.handle_request()
        server.check_orphaned(args.ppid)

    server.server_close()



"""
Collection of basic callbacks for import / export functions
You may write your own callbacks for complex cases

--------------

Export callback arguments:
1) pipe opened in "rb" mode
2) destination for callback (if applicable, can be None)
3) callback_params as dict

--------------

Import callback arguments:
1) pipe opened in "wb" mode
2) source data for callback (e.g. pandas data frame, list, file object)
3) callback_params as dict

"""

import io
import csv
import shutil


def export_to_list(pipe, dst, **kwargs):
    """
    Basic example how to export CSV stream into basic list of tuples
    """
    wrapped_pipe = io.TextIOWrapper(pipe, newline='\n', encoding='utf-8')
    reader = csv.reader(wrapped_pipe, lineterminator='\n', **kwargs)

    return [row for row in reader]


def export_to_pandas(pipe, dst, **kwargs):
    """
    Basic example how to export into Pandas DataFrame
    Custom params for "read_csv" may be passed in **kwargs
    """
    import pandas
    return pandas.read_csv(pipe, skip_blank_lines=False, **kwargs)


def export_to_file(pipe, dst):
    """
    Basic example how to export into file or file-like object opened in binary mode
    """
    if not hasattr(dst, 'write'):
        dst = open(dst, 'wb')

    shutil.copyfileobj(pipe, dst, 65535)


def import_from_iterable(pipe, src, **kwargs):
    """
    Basic example how to import from iterable object (list, dict, tuple, iterator, generator, etc.)
    """
    if not hasattr(src, '__iter__'):
        raise ValueError('Data source is not iterable')

    wrapped_pipe = io.TextIOWrapper(pipe, newline='\n', encoding='utf-8')
    writer = csv.writer(wrapped_pipe, lineterminator='\n', quoting=csv.QUOTE_NONNUMERIC, **kwargs)

    for row in src:
        writer.writerow(row)


def import_from_pandas(pipe, src, **kwargs):
    """
    Basic example how to import from Pandas DataFrame
    Custom params for "to_csv" may be passed in **kwargs
    """
    import pandas

    if not isinstance(src, pandas.DataFrame):
        raise ValueError('Data source is not pandas.DataFrame')

    # Pandas insists on outputting CSV as text
    # I could not find a good way to override it
    wrapped_pipe = io.TextIOWrapper(pipe, newline='\n', encoding='utf-8')

    return src.to_csv(wrapped_pipe, header=False, index=False, line_terminator='\n', quoting=csv.QUOTE_NONNUMERIC, **kwargs)


def import_from_file(pipe, src):
    """
    Basic example how to import from file or file-like object opened in binary mode
    """
    if not hasattr(src, 'read'):
        src = open(src, 'rb')

    shutil.copyfileobj(src, pipe, 65535)



"""
Usage: python -m pyexasol <command> <options>
"""

import argparse


parser = argparse.ArgumentParser(prog=f'python -m pyexasol')
subparsers = parser.add_subparsers(title=None, dest='command', metavar='<command>')
subparsers.required = True

subparsers.add_parser('script_output')
subparsers.add_parser('script_debug')

args = parser.parse_args()

if args.command == 'script_output':
    raise RuntimeError("'-m pyexasol script_output' is no longer available, "
                       "please use '-m pyexasol_utils.script_output' instead")

elif args.command == 'script_debug':
    raise RuntimeError("'-m pyexasol script_debug' is no longer available, "
                       "please use '-m pyexasol_utils.script_output' instead")



"""
Simple benchmark for Exasol

"""
import multiprocessing
import os
import pathlib
import pyexasol
import queue
import sys
import time


BUSY_LOOP_DELAY = 0.2
STARTUP_WORKER_DELAY = 1


class ExaBenchWorker(multiprocessing.Process):
    def __init__(self, proc_num, args, sql_text, queue):
        self.proc_num = proc_num
        self.args = args
        self.sql_text = sql_text
        self.queue = queue

        super().__init__()

    def run(self):
        C = pyexasol.connect(dsn=args.dsn, user=args.user, password=args.password, schema=args.schema, autocommit=False)
        self.queue.put(C.session_id())

        sql_comment = f'/* bench_run_id={self.proc_num} */'

        try:
            C.execute(f"ALTER SESSION SET QUERY_CACHE='OFF' {sql_comment}")

            for sql_query in sql_text.split(';'):
                formatted_sql_query = sql_query.strip("\n ")
                if len(formatted_sql_query) == 0:
                    continue

                C.execute(f"{formatted_sql_query} {sql_comment}")

            C.execute(f"COMMIT {sql_comment}")
        except pyexasol.ExaQueryError as e:
            self.queue.put(str(e))
        finally:
            C.close()


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(prog=f'python -m pyexasol_utils.bench [action]')

    parser.add_argument('action', help='Benchmark action (prepare, set1, set2, etc.), OR path to SQL file with commands')
    parser.add_argument('--dsn', default=os.environ.get('EXAHOST', ''), help='Exasol DSN (host:port)')
    parser.add_argument('--user', default=os.environ.get('EXAUID', ''), help='Exasol user name')
    parser.add_argument('--password', default=os.environ.get('EXAPWD', ''), help='Exasol password')
    parser.add_argument('--schema', default=os.environ.get('EXASCHEMA', 'PYEXASOL_BENCH'), help='Exasol schema for tests (default: PYEXASOL_BENCH)')
    parser.add_argument('--scale', default=1.0, type=float, help='Scaling factor for the amount of generated data (default: 1.0)')
    parser.add_argument('--parallel', default=1, type=int, help='Number of parallel executions')

    args = parser.parse_args()

    # Check connectivity first
    C = pyexasol.connect(dsn=args.dsn, user=args.user, password=args.password)

    # Prepare schema
    if args.action == 'prepare':
        if args.parallel > 1:
            raise ValueError("Cannot run [prepare.sql] with more than 1 parallel processes")

        if not C.meta.schema_exists(args.schema):
            C.execute("CREATE SCHEMA {schema!i}", {"schema": args.schema})

    # Get SQL text
    bench_sql_path = pathlib.Path(__file__).parent / "bench_sql" / f"{args.action}.sql"
    bench_sql_path.resolve()

    if not bench_sql_path.is_file():
        bench_sql_path = pathlib.Path(args.action)

        if not bench_sql_path.is_file():
            raise ValueError("Action argument is neither a standard action, nor path to SQL file")

    sql_text = C.format.format(bench_sql_path.read_text(), scale=args.scale)

    # Show startup message
    if args.action == 'prepare':
        print(f"Running [{bench_sql_path.name}] with scale factor of [{args.scale}]", file=sys.stderr)
    else:
        print(f"Running [{bench_sql_path.name}] in [{args.parallel}] parallel processes", file=sys.stderr)

    # Start worker processes
    process_pool = list()
    message_queue = multiprocessing.Queue()

    for i in range(args.parallel):
        proc = ExaBenchWorker(i, args, sql_text, message_queue)
        proc.start()

        process_pool.append(proc)
        time.sleep(STARTUP_WORKER_DELAY)

    while len(process_pool) > 0:
        time.sleep(BUSY_LOOP_DELAY)

        for p in process_pool:
            if not p.is_alive():
                p.join()
                process_pool.remove(p)

                print(f"Finished worker {p.proc_num}", file=sys.stderr)

    # Display errors
    error_cnt = 0
    session_id_list = list()

    while True:
        try:
            msg = str(message_queue.get_nowait())

            if msg.isnumeric():
                session_id_list.append(msg)
            else:
                print(msg, file=sys.stderr)
                error_cnt = error_cnt + 1
        except queue.Empty:
            break

    # Flush statistics
    C.execute("FLUSH STATISTICS")

    if error_cnt > 0:
        print(f"Error count: {error_cnt}", file=sys.stderr)
        exit(0)

    # Output SQL to get results
    run_id_regex_replace = r"REGEXP_REPLACE(sql_text, '(?s)^.*\/\* bench_run_id=(\d+) \*\/.*$', '\1')"

    print(f"""
Durations SQL:
SELECT session_id
    , stmt_id
    , CAST({run_id_regex_replace} AS DECIMAL(9,0)) AS run_id
    , command_name
    , duration
    , avg(duration) OVER (PARTITION BY stmt_id) AS avg_duration
    , sql_text
FROM "$EXA_AUDIT_SQL"
WHERE session_id IN ({','.join(session_id_list)})
    AND stmt_id > 1
ORDER BY 2,3;
    """, file=sys.stderr)

    print(f"""
Profiling details SQL:
SELECT session_id
    , stmt_id
    , command_name
    , part_id
    , iproc
    , duration
    , part_name
    , part_info
    , object_schema
    , object_name
    , object_rows
    , in_rows
    , out_rows
    , remarks
FROM "$EXA_PROFILE_DETAILS_LAST_DAY"
WHERE session_id = {session_id_list[-1]}
    AND stmt_id > 1
ORDER BY 2,4,5;
    """, file=sys.stderr)

    C.close()
